%%%-------------------------------------------------------------------
%%% @author Piotr Duleba
%%% @copyright (C) 2021 ACK CYFRONET AGH
%%% This software is released under the MIT license
%%% cited in 'LICENSE.txt'.
%%% @end
%%%-------------------------------------------------------------------
%%% @doc
%%% This file provides functions for rpc test modules.
%%% @end
%%%-------------------------------------------------------------------
-module(test_rpc).
-author("Piotr Duleba").

-export([call/5, call/6]).

-type service() :: op_worker | oz_worker | onepanel.

%%%===================================================================
%%% API functions
%%%===================================================================


-spec call(service(), oct_background:node_selector(), module(), atom(), [term()]) ->
    term() | no_return().
call(Service, NodeSelector, Module, Function, Args) ->
    call(Service, NodeSelector, Module, Function, Args, infinity).


-spec call(service(), oct_background:node_selector(), module(), atom(), [term()], timeout()) ->
    term() | no_return().
call(Service, NodeSelector, Module, Function, Args, Timeout) ->
    Node = case is_known_node(NodeSelector) of
        true ->
            NodeSelector;
        false ->
            lists_utils:random_element(get_service_nodes(Service, NodeSelector))
    end,

    try
        erpc:call(Node, Module, Function, Args, Timeout)
    catch Type:Reason when Type =:= exit; Type =:= error ->
        ct:pal(
            "Test RPC on node ~p failed!~n"
            "MFA: ~w:~w(~w)~n"
            "Reason: ~w",
            [Node, Module, Function, Args, Reason]
        ),
        error({test_rpc_failed, Node, Module, Function, Reason})
    end.


%%%===================================================================
%%% Helpers
%%%===================================================================


%% @private
-spec is_known_node(oct_background:node_selector()) -> boolean().
is_known_node(NodeSelector) ->
    lists:member(NodeSelector, nodes(known)).


%% @private
-spec get_service_nodes(op_worker | oz_worker | onepanel, oct_background:node_selector()) -> [node()].
get_service_nodes(Service, NodeSelector) ->
    case Service of
        op_worker -> oct_background:get_provider_nodes(NodeSelector);
        oz_worker -> oct_background:get_zone_nodes();
        onepanel -> oct_background:get_service_panels(NodeSelector)
    end.
